# PWBus - KafkaFlowIn & KafkaFlowOut Classes
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Wed Dec 11 06:45:11 -03 2019

from pykafka.common import OffsetType
from pykafka.exceptions import SocketDisconnectedError, LeaderNotAvailable
import traceback

from pwbus.commons.logging import *
from pwbus.flows._flow import FlowIn, FlowOut
from pwbus.connectors.kafka import KafkaConnectorIn, KafkaConnectorOut

try:
    from pykafka.rdkafka import _rd_kafka  # noqa
    USE_RDKAFKA = True
    log_debug("USE_RDKAFKA Enabled!")
except ImportError:
    USE_RDKAFKA = False  # C extension not built
    log_debug("⚠️  USE_RDKAFKA Disabled!")

# KafkaFlowIn
#
#


class KafkaFlowIn(FlowIn):

    def __init__(self, request=None, host=None, port=None):
        super().__init__('in', KafkaConnectorIn(host=host, port=port))
        self.host = host
        self.port = port

    # KafkaFlowIn.receive
    #
    def receive(self, timeout=60):
        DEBUG = self.isDebugEnabled()
        resource_name = self.getRecvFrom()

        if DEBUG:
            log_debug(
                f'KafkaFlowIn.receive - Started to [{resource_name}] format={self.getPayloadFormat()}')

        try:
            if DEBUG:
                log_debug("KafkaFlowIn.receive.message - Getting connection")

            with self.getConnector().getConnectionPool().item() as connection:

                topic = connection.topics[resource_name]

                # pooling for wait messages
                while True:
                    consumer = topic.get_simple_consumer(
                        consumer_group=resource_name.encode(),
                        queued_max_messages=1,
                        consumer_timeout_ms=int(timeout*1000),
                        auto_commit_enable=True,
                        fetch_wait_max_ms=50
                    )
                    message = consumer.consume()
                    if message is not None:
                        payload_in = ''.join(message.value.decode('utf-8'))

                        consumer.commit_offsets()

                        if DEBUG:
                            log_debug(
                                f'↘️  KafkaFlowIn.receive.message - Message received successfully from [{resource_name}] - payload_in [{payload_in}]')
                        return payload_in

        except KeyboardInterrupt:
            log_debug('KafkaFlowIn.receive.message - Interrupted by user')
            self.stop()

        except:
            log_error(
                traceback, f'KafkaFlowIn.receive.message - Failed on receive message from [{self.getRecvFrom()}]')

        return None


# KafkaFlowOut
#
#
class KafkaFlowOut(FlowOut):

    def __init__(self, host=None, port=None):
        super().__init__('out', KafkaConnectorOut(host=host, port=port))
        self.host = host
        self.port = port

    # KafkaFlowOut.send
    #
    def send(self):
        DEBUG = self.isDebugEnabled()

        resource_name = self.getSendTo()
        payload_out = self.getMessage().getPayloadStream()

        if DEBUG:
            log_debug(
                f'KafkaFlowOut.send - Started to [{resource_name}] - payload_out [{payload_out}]')

        try:
            with self.getConnector().getConnectionPool().item() as connection:

                topic = connection.topics[resource_name]

                with topic.get_producer(
                        use_rdkafka=USE_RDKAFKA,
                        min_queued_messages=1,
                        linger_ms=0) as producer:
                    try:
                        producer.produce(payload_out)

                    except (SocketDisconnectedError, LeaderNotAvailable) as e:
                        print(e)
                        producer = topic.get_producer()
                        producer.stop()
                        producer.start()
                        producer.produce(payload_out)

                    if DEBUG:
                        log_debug(
                            f'↗️  KafkaFlowOut.send - Message sent successfully to [{resource_name}]')
                    return True
        except:
            log_error(
                traceback, f'KafkaFlowOut.send - Failed to send the message to [{resource_name}]')
            traceback.print_exc()

        return False



# PWBus - EngineMonitorEvent Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Mon Dec 16 20:32:30 -03 2019

from datetime import datetime
from json import dumps
import traceback

from pwbus.commons.logging import *
from pwbus.connectors.redis import RedisConnector

# EngineMonitorEvent - Get engine monitor channel settings
#
#


class EngineMonitorEvent(RedisConnector):

    def __init__(self, connection=None):
        if not connection:
            super().__init__()
            with self.getConnectionPool().item() as conn:
                self.connection = conn
        else:
            self.connection = connection

    # EngineMonitorEvent.pushEvent
    #
    def pushEvent(self, value):
        try:
            value.update(
                {"datetime": datetime.now().strftime('%Y-%m-%d %H:%M:%S')})
            self.connection.lpush('pwbus_monitor_events', dumps(value))
        except Exception as e:
            print('EngineMonitorEvent.send - Failed on push event', str(e))

    # EngineMonitorEvent.fetchEvent
    #
    def fetchEvent(self):
        try:
            event = self.connection.rpop('pwbus_monitor_events')
            if event:
                return ''.join(event.decode('utf-8'))
        except Exception as e:
            print('EngineMonitorEvent.send - Failed on push event', str(e))

    # EngineMonitorEvent.setData
    #
    async def setData(self, name, value):
        try:
            self.connection.set(name, dumps(value))
        except Exception as e:
            print('EngineMonitorEvent.setData - Failed on set data', str(e))

    # EngineMonitorEvent.getData
    #
    def getData(self, name):
        try:
            data = self.connection.get(name)
            if data:
                if isinstance(data, int):
                    return str(data)
                else:
                    return ''.join(data.decode('utf-8'))
        except Exception as e:
            print('EngineMonitorEvent.getData - Failed on get data', str(e))

    # EngineMonitorEvent.setValue
    #
    def setValue(self, name, value):
        try:
            self.connection.set(name, value)
        except Exception as e:
            print('EngineMonitorEvent.setValue - Failed on set value', str(e))

    # EngineMonitorEvent.incrValue
    #
    def incrValue(self, name):
        try:
            data = self.connection.incr(name, amount=1)
            if data:
                return str(data)
        except Exception as e:
            print('EngineMonitorEvent.incrValue - Failed on increment value', str(e))
            print(str(e))



#!/usr/bin/env python3

# PWBus - Main
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Tue Nov 19 11:17:54 -03 2019

__author__ = "fabio.szostak@perfweb.com.br"
__version__ = "0.1.28"

import sys
import os.path
import getopt

from pwbus.engines.engine_manager import EngineManager


def main():
    argv = sys.argv[1:]

    channel = None
    engine = None
    command = None
    registry_file = None

    help_message = f'\npwbus -f <registry-file> [-v] [-e <engine>] [-c <channel>] start\n\nPWBUS Core - Version {__version__}\n\n      -v  version\n      -f  <registry-file>\n      --registry=<registry-config-file>\n      -e <redis|socket>\n      ----engine=<redis|socket>\n\n      -c <channel_id>\n      --channel=channel_id\n\n      start the pwbus server\n'

    try:
        opts, args = getopt.getopt(argv, "hve:c:f:", ["start|registry=|channel=|engine="])
    except getopt.GetoptError:
        print(help_message)
        sys.exit(2)

    for arg in args:
        if arg == "start":
            command = arg

    for opt, arg in opts:
        if opt == '-h':
            print(help_message)
            sys.exit()
        elif opt == '-v':
            print(__version__)
            sys.exit()
        elif opt in ("-e", "--engine"):
            engine = arg
        elif opt in ("-c", "--channel"):
            channel = arg
        elif opt in ("-f", "--registry-file"):
            registry_file = arg

    try:
        if not registry_file:
            if os.path.isfile("./etc/pwbus_registry.json"):
                registry_file = "./etc/pwbus_registry.json"
            else:
                print("\nError: Registry file is not defined")
        else:
            if command:
                engine_manager = EngineManager(registry_file)
                engine_manager.action(command, engine, channel)
                sys.exit(0)

    except:
        sys.exit(1)

    print(help_message)


# __main__
#
if __name__ == "__main__":
    main()



# PWBus - KafkaConnector Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Wed Dec 11 06:44:08 -03 2019

from pykafka import KafkaClient
import traceback

from pwbus.commons.logging import *
from pwbus.connectors._connector import _Connector

# KafkaConnectorIn
#
#


class KafkaConnector(_Connector):

    def __init__(self, host='kafka', port=9092):
        super().__init__('kafka', host, port)

        self.setConnectionPool(
            self.create_connection,
            max_size=10,
            max_usage=1000,
            idle=300,
            ttl=600
        )

    # KafkaConnector.create_connection
    #
    def create_connection(self):
        try:
            return KafkaClient(hosts=f'{self.host}:{self.port}')
        except:
            log_error(
                traceback, 'KafkaConnector - Error on try connect to KAFKA server')


# KafkaConnectorIn
#
#
class KafkaConnectorIn(KafkaConnector):

    def __init__(self, host, port):
        super().__init__(host=host, port=port)

# KafkaConnectorOut
#
#


class KafkaConnectorOut(KafkaConnector):

    def __init__(self, host, port):
        super().__init__(host=host, port=port)



# PWBus - cprint - color print
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Tue Apr 28 18:17:12 -03 2020

COLORS = True

bcolors = {
    "black":          u'\u001b[30m',
    "red":            u'\u001b[31m',
    "green":          u'\u001b[32m',
    "yellow":         u'\u001b[33m',
    "blue":           u'\u001b[34m',
    "magenta":        u'\u001b[35m',
    "cyan":           u'\u001b[36m',
    "white":          u'\u001b[37m',
    "bright_black":   u'\u001b[40m\u001b[1m',
    "bright_red":     u'\u001b[41m\u001b[1m',
    "bright_green":   u'\u001b[42m\u001b[1m',
    "bright_yellow":  u'\u001b[43m\u001b[1m',
    "bright_blue":    u'\u001b[44m\u001b[1m',
    "bright_magenta": u'\u001b[45m\u001b[1m',
    "bright_cyan":    u'\u001b[46m\u001b[1m',
    "bright_white":   u'\u001b[47m\u001b[1m',
}

fcolors = {
    "reset":          u'\u001b[0m',
    "bold":           u'\u001b[1m',
    "black":          u'\u001b[30m',
    "red":            u'\u001b[31m',
    "green":          u'\u001b[32m',
    "yellow":         u'\u001b[33m',
    "blue":           u'\u001b[34m',
    "magenta":        u'\u001b[35m',
    "cyan":           u'\u001b[36m',
    "white":          u'\u001b[37m',
    "bright_black":   u'\u001b[30m\u001b[1m',
    "bright_red":     u'\u001b[31m\u001b[1m',
    "bright_green":   u'\u001b[32m\u001b[1m',
    "bright_yellow":  u'\u001b[33m\u001b[1m',
    "bright_blue":    u'\u001b[34m\u001b[1m',
    "bright_magenta": u'\u001b[35m\u001b[1m',
    "bright_cyan":    u'\u001b[36m\u001b[1m',
    "bright_white":   u'\u001b[37m\u001b[1m',
}

thread_colors = [
    "white",
    "green",
    "cyan",
    "bright_blue",
    "bright_green",
    "blue",
    "magenta",
    "yellow",
    "bright_green",
    "bright_cyan",
    "bright_magenta"
]

thread_list = dict()


def cprint(data, color='yellow', bcolor='black', enable=COLORS):
    print(f"{bcolors[bcolor]}{fcolors[color]}{data}{fcolors['reset']}" if enable else data, end='')


def cprint_rainbow(data, key='thread', bcolor='black', enable=COLORS):
    global thread_list
    global thread_colors

    fcolor = thread_list[key] if key in thread_list else None
    if not fcolor:
        count = len(thread_list.keys())
        if count < 10:
            fcolor = thread_colors[count]
        else:
            fcolor = 'white'
        thread_list[key] = fcolor

    print(f"{bcolors[bcolor]}{fcolors[fcolor]}{data}{fcolors['reset']}" if enable else data, end='')



# PWBus - SqsClient Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sun Nov 17 20:10:24 -03 2019

import boto3
from json import dumps, loads, load
from time import sleep
import traceback

from pwbus.commons.logging import *
from pwbus.clients.client import Client
from pwbus.connectors.sqs import SqsConnector

# SqsClient
#
#


class SqsClient(SqsConnector, Client):

    def __init__(self, host=None, port=None):
        super().__init__()

    # SqsClient.get
    #
    def get(self, resource_name, correlation_id, wait=0.05, retries=20):

        log_debug(
            f'🔎 SqsClient.get - Retrieving response from [{resource_name}] with correlation_id [{correlation_id}] wait={wait}ms retries={retries}')
        try:
            with self.getConnectionPool().item() as connection:

                received_messages = connection.receive_message(
                    QueueUrl=resource_name,
                    MaxNumberOfMessages=5,
                    MessageAttributeNames=[
                        'All'
                    ],
                    WaitTimeSeconds=int(wait*retries),
                    VisibilityTimeout=1
                )

                if 'Messages' in received_messages:
                    for message in received_messages['Messages']:
                        if correlation_id == message['MessageAttributes'].get('Pwbus-Correlation-Id').get('StringValue'):
                            response = loads(message['Body'])

                            connection.delete_message(
                                QueueUrl=resource_name,
                                ReceiptHandle=message['ReceiptHandle']
                            )

                            log_debug(
                                f'SqsClient.get - Response found - correlation_id [{correlation_id}]]')
                            return self.clear_header(response)

                log_debug(
                    f'SqsClient.get - Response not found - correlation_id [{correlation_id}]')

        except:
            log_error(
                traceback, f'SqsClient.get - Failed to get message by correlation_id')

        return None



# PWBus - SqsEngine Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Dec  7 09:46:09 -03 2019

import traceback

from pwbus.commons.logging import *
from pwbus.engines._engine import _Engine
from pwbus.engines.sqs.engine_thread import SqsEngineThread


class SqsEngine(_Engine):

    def __init__(self):
        super().__init__(
            name='sqs'
        )

    # SqsEngine.serve
    #
    def serve(self, channel_registry):
        return self.run(SqsEngineThread, channel_registry)



# PWBus - RegistryLoader Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Tue Nov 19 21:05:19 -03 2019

from os import getenv
import redis
from json import dumps, load
import traceback

from pwbus.commons.logging import *
from pwbus.connectors.redis import RedisConnector

# Registry loader
#
#


class RegistryLoader(RedisConnector):
    def __init__(self, registry_file):
        RedisConnector.__init__(self)
        self.registry_file = registry_file

    # RegistryLoader.load
    #
    def load(self):
        try:
            with open(self.registry_file) as json_file:
                registry_list = load(json_file)

            with self.getConnectionPool().item() as connection:
                for item in registry_list:
                    connection.delete(f'wpbus.registry.{item["channel"]}')
                    connection.set(
                        f'wpbus.registry.{item["channel"]}', dumps(item))

            log_debug('PWBUS Registry - Loaded successfully')

        except:
            log_error(
                traceback, 'RegistryLoader.load - Error trying to load the registry')
            log_exit('RegistryLoader.load - exited')

    # RegistryLoader.getRunnableEngines
    #
    def getRunnableEngines(self):
        engines = []

        try:
            with open(self.registry_file) as json_file:
                registry_list = load(json_file)

            for channel_entry in registry_list:
                if channel_entry["engine.enabled"]:
                    if channel_entry["flow.in.resource_type"] != 'http' and \
                            not channel_entry["flow.in.resource_type"] in engines:
                        engines.append(channel_entry["flow.in.resource_type"])

            log_debug(f'Registry.getRunnableEngines - Engines {engines}')
            return engines

        except:
            log_error(
                traceback, 'Registry.getRunnableEngines - Error getting engines')
            log_exit('RegistryLoader.getRunnableEngines - exited')

    # RegistryLoader.getRunnableChannels
    #
    def getRunnableChannels(self, engine):
        channels = []

        try:
            with open(self.registry_file) as json_file:
                registry_list = load(json_file)

            for channel_entry in registry_list:
                if channel_entry["flow.in.resource_type"] == engine:
                    if not channel_entry['channel'] in channels:
                        channels.append(channel_entry['channel'])

            log_debug(f'Registry.getRunnableChannels - Channels {channels}')
            return channels

        except:
            log_error(
                traceback, 'Registry.getRunnableChannels - Error getting channels')
            log_exit('RegistryLoader.getRunnableChannels - exited')



# PWBus - _EngineThread Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

import threading
import traceback

from pwbus.commons.logging import *
from pwbus.engines.engine_dispatch import EngineDispatch

# _EngineThread
#
#


class _EngineThread(threading.Thread):

    def __init__(self, thread_id, name, channel_registry):
        threading.Thread.__init__(self)
        self.thread_id = thread_id
        self.name = name
        self.channel_registry = channel_registry
        self.channel = channel_registry["channel"]
        self.this_thread = threading.currentThread()

    # _EngineThread.stop
    #
    def stop(self):
        self.this_thread.do_run = False
        self.this_thread.join()

    # _EngineThread.isRunning
    #
    def isRunning(self):
        return getattr(self.this_thread, "do_run", True)

    # _EngineThread.dispatch
    #
    def dispatch(self, request=None):
        try:
            engine_dispatch = EngineDispatch()
            return engine_dispatch.route(self.channel_registry, request)
        except KeyboardInterrupt:
            log_debug(
                f'_EngineThread.run - Stopping thread for channel [{self.channel}]')
            self.stop()
            return None
        except:
            log_error(
                traceback, f'_EngineThread.run - Error thread execution for channel [{self.channel}]')

    # _EngineThread.run
    #

    def run(self):
        if not self.channel_registry:
            log_fatal(
                f'_EngineThread.run - channel [{self.channel}] not found')

        DEBUG = self.channel_registry['engine.debug']

        if DEBUG:
            log_debug(
                f'_EngineThread.run - Starting thread for channel [{self.channel}]')
            #log_debug_var(f'_EngineThread.run - channel [{self.channel}]', self.channel_registry)

        while True:
            try:
                if not self.isRunning():
                    if DEBUG:
                        log_debug(
                            f'_EngineThread.run - [{self.channel}] stopped ')
                    break

                response_flow = self.dispatch()

                if not response_flow:
                    log_debug('_EngineThread.run - Failed on dispatch message')

            except KeyboardInterrupt:
                self.stop()

            except:
                log_error(
                    traceback, '_EngineThread.run - Flow interrupted with errors')

        log_debug(f'_EngineThread.run - Stopping thread [{self.name}]')



# PWBus - RedisClient Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sun Nov 17 20:10:24 -03 2019

import redis
from json import dumps, loads, load
from time import sleep
from datetime import datetime
import traceback

from pwbus.commons.logging import *
from pwbus.clients.client import Client
from pwbus.connectors.redis import RedisConnector

# RedisClient
#
#
class RedisClient(RedisConnector, Client):

    def __init__(self, host=None, port=None):
        super().__init__(host=host, port=port)

    # RedisClient.get
    #
    def get(self, resource_name, correlation_id, wait=0.00, retries=5):
        log_debug(
            f'🔎 RedisClient.get - Retrieving response from [{resource_name}] with correlation_id [{correlation_id}] wait={wait}ms retries={retries}')
        try:
            with self.getConnectionPool().item() as connection:
                response = None

                log_debug(
                    f'🔎 RedisClient.get - Connection retrieved from pool for [{resource_name}]')

                start = getMillis()
                scan_result = connection.sscan_iter(name=resource_name, match=f'*Pwbus-Correlation-Id*{correlation_id}*')
                end = getMillis()

                if scan_result is not None:
                    scan_time = end - start
                    response = scan_result.decode("utf-8")
                    connection.srem(resource_name, response)
                    connection.rpop(
                        f'{resource_name}_helper'
                    )
                    response = loads(response)
                    end_racing = getMillis()
                    log_debug(
                        f'RedisClient.get - Response found - correlation_id [{correlation_id}] retry [{retry+1}] wait [{wait}] scan-time [{scan_time}ms]')
                    return self.clear_header(response)

                return None

        except:
            log_error(
                traceback, f'RedisClient.get - Failed to get message by correlation_id')

        return None
